// Copyright (c) 1999-2011  Pavel Rychly

#include <config.hh>
#include <finlib/excep.hh>
#include <finlib/seek.hh>
#include <finlib/huffman.hh>
#include <cstring>
#include <climits>

using namespace std;

static int reverse_bits (int x, int len)
{
    int r = 0;
    while (len--) {
	r <<= 1;
	if (x & 1)
	    r |= 1;
	x >>= 1;
    }
    return r;
}


huffman_data::huffman_data (const char *filename, int read_write)
{
    FILE *f = fopen (filename, "r");
    if (f == NULL) throw FileAccessError (filename, "huffman_data: fopen");
    fread (&num, sizeof(num), 1, f);
    fread (&maxlen, sizeof(maxlen), 1, f);

    firstcode = new CodeType [maxlen];
    fread (firstcode, sizeof(CodeType), maxlen, f);

    if (read_write & rw_READ) {
	symidx = new int [maxlen];
	fread (symidx, sizeof(int), maxlen, f);
    } else {
	symidx = 0;
	fseeko (f, sizeof(int) * maxlen, SEEK_CUR);
    }

    if (read_write & rw_READ) {
	symbol = new int [num];
	fread (symbol, sizeof(CodeType), num, f);
    } else {
	symbol = 0;
	fseeko (f, sizeof(CodeType) * num, SEEK_CUR);
    }

    if (read_write & rw_WRITE) {
	lens = new char [num];
	fread (lens, sizeof(char), num, f);
    } else
	lens = 0;
    fclose (f);

    if (read_write & rw_WRITE) {
	codes = new CodeType [num];
	CodeType *next = new CodeType [maxlen];
	memcpy (next, firstcode, sizeof (CodeType) * maxlen);
	for (int i=0; i < num; i++) {
	    int l = lens[i];
	    codes [i] = reverse_bits (next[l], l);
	    //codes [i] = next[l];
	    next [l] ++;
	}
	delete next;
    } else
	codes = 0;
}

huffman_data::~huffman_data ()
{
    if (firstcode) delete firstcode;
    if (symbol) delete symbol;
    if (symidx) delete symidx;
    if (lens) delete lens;
    if (codes) delete codes;
}

bool huffman_data::store (char *filename)
{
    FILE *f = fopen (filename, "w");
    if (!f) return false;
    fwrite (&num, sizeof(num), 1, f);
    fwrite (&maxlen, sizeof(maxlen), 1, f);

    fwrite (firstcode, sizeof(CodeType), maxlen, f);
    fwrite (symidx, sizeof(int), maxlen, f);
    fwrite (symbol, sizeof(CodeType), num, f);
    fwrite (lens, sizeof(char), num, f);
    fclose (f);
    return true;
}

huffman_data::huffman_data (int fnum, int *freqs)
    : num (fnum), maxlen (sizeof (CodeType) *8)
{
    firstcode = new CodeType [maxlen];
    symidx = new int [maxlen];
    symbol = new int [num];
    lens = new char [num];
    maxlen = generate_data (freqs);
    
    codes = new CodeType [num];
    CodeType *next = new CodeType [maxlen];
    memcpy (next, firstcode, sizeof (CodeType) * maxlen);
    for (int i=0; i < num; i++) {
	int l = lens[i];
	codes [i] = reverse_bits (next[l], l);
	symbol [symidx[l] + next[l] - firstcode[l]] = i;
	next [l] ++;
    }
    delete next;
}

int huffman_data::generate_data (int *freqs)
{
    int HNum, i;
    int *heap = new int [2*num];
    int *lencount = new int [maxlen];

    /* Initialise the leaves */
    memcpy (&heap[num], freqs, num * sizeof (int));

    /* Initialise the pointers to the leaves */
    for (i = 0; i < num; i++)
	heap [i] = num +i;

    /* Reorganise the pointers so that it is a heap */
    HNum = num;
    for (i = HNum / 2; i > 0; i--) {
	int curr = i;
	int child = curr * 2;
	while (child <= HNum) {
	    if (child < HNum && heap[heap[child]] < heap[heap[child - 1]])
		child++;
	    if (heap[heap[curr - 1]] > heap[heap[child - 1]]) {
		int temp = heap[child - 1];
		heap[child - 1] = heap[curr - 1];
		heap[curr - 1] = temp;
		curr = child;
		child = 2 * curr;
	    } else
		break;
	}
    }

    /* Form the huffman tree */
    while (HNum > 1) {
	int pos[2];

	for (i = 0; i < 2; i++)	{
	    int curr = 1;
	    int child = 2;
	    pos[i] = heap[0];
	    heap[0] = heap[--HNum];
	    while (child <= HNum) {
		if (child < HNum && heap[heap[child]] < heap[heap[child - 1]])
		    child++;
		if (heap[heap[curr - 1]] > heap[heap[child - 1]]) {
		    int temp = heap[child - 1];
		    heap[child - 1] = heap[curr - 1];
		    heap[curr - 1] = temp;
		    curr = child;
		    child = 2 * curr;
		} else
		    break;
	    }
	}

	heap[HNum + 1] = heap[pos[0]] + heap[pos[1]];
	heap[pos[0]] = heap[pos[1]] = HNum + 1;

	heap[HNum] = HNum + 1;

	HNum++;
	int curr = HNum;
	int parent = curr >> 1;
	while (parent && heap[heap[parent - 1]] > heap[heap[curr - 1]]) {
	    int temp = heap[parent - 1];
	    heap[parent - 1] = heap[curr - 1];
	    heap[curr - 1] = temp;
	    curr = parent;
	    parent = curr >> 1;
	}
    }


    /* Calculate the code lens */
    heap[0] = ULONG_MAX;
    heap[1] = 0;
    for (i = 2; i < num * 2; i++)
	heap[i] = heap[heap[i]] + 1;

    int maxcodelen = 0;
    /* Set the code length of each leaf in the huffman tree */
    for (i = 0; i < num; i++) {
	int codelen = heap[i + num];
	lens[i] = codelen;
	lencount[codelen]++;
	if (codelen > maxcodelen)
	    maxcodelen = codelen;
    }
    maxcodelen++;

    /* Calculate the current codes for each different code length */
    firstcode[maxcodelen] = 0;
    for (i = maxcodelen - 1; i>=0; i--)
	firstcode[i] = (firstcode[i + 1] + lencount[i + 1]) >> 1;

    int sum = 0;
    for (i = 0; i < maxcodelen; i++) {
	symidx [i] = sum;
	sum += lencount [i];
    }

    delete heap;
    delete lencount;
    return maxcodelen;
}

